r"""

Filters
#######

**Highlight Features of Interest**

This module contains a variety of functions for altering images based on
the structural characteristics, such as pore sizes.  A definition of a
*filter* is a function that returns an image the shape as the original
image, but with altered values.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    filters.apply_chords
    filters.apply_chords_3D
    filters.apply_padded
    filters.chunked_func
    filters.distance_transform_lin
    filters.fftmorphology
    filters.fill_blind_pores
    filters.find_disconnected_voxels
    filters.find_dt_artifacts
    filters.find_peaks
    filters.flood
    filters.flood_func
    filters.hold_peaks
    filters.local_thickness
    filters.nphase_border
    filters.nl_means_layered
    filters.porosimetry
    filters.prune_branches
    filters.reduce_peaks
    filters.region_size
    filters.size_to_seq
    filters.size_to_satn
    filters.seq_to_satn
    filters.snow_partitioning
    filters.snow_partitioning_n
    filters.snow_partitioning_parallel
    filters.trim_disconnected_blobs
    filters.trim_extrema
    filters.trim_floating_solid
    filters.trim_nearby_peaks
    filters.trim_nonpercolating_paths
    filters.trim_saddle_points
    filters.trim_small_clusters


"""

from ._funcs import apply_chords
from ._funcs import apply_chords_3D
from ._funcs import apply_padded
from ._funcs import chunked_func
from ._funcs import distance_transform_lin
from ._funcs import fill_blind_pores
from ._funcs import find_disconnected_voxels
from ._funcs import find_dt_artifacts
from ._funcs import flood
from ._funcs import flood_func
from ._funcs import hold_peaks
from ._funcs import local_thickness
from ._funcs import nphase_border
from ._funcs import porosimetry
from ._funcs import prune_branches
from ._funcs import region_size
from ._funcs import trim_disconnected_blobs
from ._funcs import trim_extrema
from ._funcs import trim_floating_solid
from ._funcs import trim_nonpercolating_paths
from ._funcs import trim_small_clusters
from ._snows import snow_partitioning
from ._snows import snow_partitioning_n
from ._snows import snow_partitioning_parallel
from ._snows import find_peaks
from ._snows import reduce_peaks
from ._snows import trim_nearby_peaks
from ._snows import trim_saddle_points
from ._size_seq_satn import size_to_seq
from ._size_seq_satn import size_to_satn
from ._size_seq_satn import seq_to_satn
from ._nlmeans import nl_means_layered
from ._fftmorphology import fftmorphology
from . import imagej
from ._ibip import ibip
from ._ibip import find_trapped_regions
from ._ibip_gpu import ibip_gpu


import numpy as np
import porespy as ps
import scipy.ndimage as spim
import skimage
import skfmm
from skimage.morphology import disk, square, ball, cube
from porespy.tools import get_tqdm
from porespy.filters import fftmorphology
from porespy import settings

tqdm = get_tqdm()


def geometrical_tortuosity(im, axis=0):
    r"""
    Calculate geometrical tortuosity across an image

    Parameters
    ----------
    im : ndarray
        The image for which geometrical tortuosity is found
    axis: int
        The axis across which to find the geometrical tortuosity

    Returns
    -------
    tau : float
        The computed tortuosity of the image along the given axis

    """
    im = np.swapaxes(im, 0, axis)
    im = ps.filters.fill_blind_pores(im)
    if np.ndim(im) == 2:
        padding = [[10, 10], [0, 0]]
        structure = disk(1)
    if np.ndim(im) == 3:
        padding = [[10, 10], [0, 0], [0, 0]]
        structure = ball(1)
    im2 = np.pad(im, padding, "maximum")
    sk1 = skimage.morphology.skeletonize_3d(im2) > 0
    sk1 = ps.tools.unpad(sk1, padding)
    imlen = sk1.shape[0]
    sk = fftmorphology(sk1, strel=structure, mode='dilation')

    inlets = np.zeros_like(sk, dtype=bool)
    inlets[0, ...] = True
    inlets = inlets*sk
    labels, n = spim.label(inlets)
    mins = []
    inl = spim.find_objects(labels)
    for i in tqdm(range(n), **settings.tqdm):

        phi = np.ones_like(sk)
        phi[inl[i]] = 0

        v = (sk)*1.0
        t = skfmm.travel_time(phi=phi, speed=v)

        out = sk1*t
        out = out.data
        out2 = out[-1, ...]
        out3 = []
        for ii in range(np.shape(out2)[0]):
            if np.ndim(im) == 2:
                if out2[ii] != 0.0 and out2[ii] != 1.0:
                    out3.append(out2[ii])
            if np.ndim(im) == 3:
                for iii in range(np.shape(out2)[1]):
                    if out2[ii, iii] != 0.0 and out2[ii, iii] != 1.0:
                        out3.append(out2[ii, iii])
        if out3 != []:
            mins.append(min(out3))

    avgmin = sum(mins)/len(mins)
    tortuosity = avgmin/imlen
    im = np.swapaxes(im, 0, axis)
    return tortuosity


def geometrical_tortuosity_points(im, axis=0):
    r"""
    Calculate geometrical tortuosity of an image within

    Parameters
    ----------
    im : ndarray
        The image for which geometrical tortuosity is found
    axis: int
        The axis across which to find the geometrical tortuosity

    """
    im = np.swapaxes(im, 0, axis)
    im = ps.filters.fill_blind_pores(im)
    if np.ndim(im) == 2:
        padding = [[10, 10], [0, 0]]
        structure = disk(1)
        footprint = disk(5)
        weights = square(3)
    if np.ndim(im) == 3:
        padding = [[10, 10], [0, 0], [0, 0]]
        structure = ball(1)
        footprint = ball(5)
        weights = cube(3)
    im2 = np.pad(im, padding, "maximum")
    sk1 = skimage.morphology.skeletonize_3d(im2) > 0
    sk1 = ps.tools.unpad(sk1, padding)
    sk = fftmorphology(sk1, strel=structure, mode='dilation')

    dns = ps.dns.tortuosity(im=sk, axis=0, return_im=True,
                            solver_family='pypardiso')
    mx = spim.maximum_filter(dns.image, footprint=footprint)*sk
    result = (dns.image < mx)
    inters = spim.convolve(result.astype(int)*sk1, weights=weights)*sk1
    inters = spim.binary_dilation((inters == 4), structure=weights)
    inters = ps.filters.reduce_peaks(inters)

    if np.ndim(im) == 2:
        labinters, num = spim.label(inters)
        locx, locy = np.nonzero(inters)
        tort = np.zeros([num, num])
        inl = spim.find_objects(labinters)
        for i in range(num):
            phi = np.ones_like(sk)
            phi[inl[i]] = 0
            v = (sk)*1.0
            t = skfmm.travel_time(phi=phi, speed=v)
            a = np.asarray([locx[i], locy[i]])
            for ii in range(num):
                b = np.asarray([locx[ii], locy[ii]])
                outl = spim.find_objects(labinters)[ii]
                out2 = t.data[outl]
                dist = np.linalg.norm(a-b)
                if i != ii:
                    tort[i, ii] = out2/dist
        tortavg = sum(sum(tort))/(num*(num-1))

    if np.ndim(im) == 3:
        labinters, num = spim.label(inters)
        locx, locy, locz = np.ndarray.nonzero(inters)
        tort = np.zeros([num, num, num])
        inl = spim.find_objects(labinters)
        for i in range(num):
            phi = np.ones_like(sk)
            phi[inl[i]] = 0
            v = (sk)*1.0
            t = skfmm.travel_time(phi=phi, speed=v)
            a = np.asarray([locx[i], locy[i], locz[i]])
            for ii in range(num):
                b = np.asarray([locx[ii], locy[ii], locz[ii]])
                outl = spim.find_objects(labinters)[ii]
                out2 = np.asarray(t.data[outl])
                dist = np.linalg.norm(a-b)

                if i != ii:
                    tort[i, ii] = out2/dist
        tortavg = sum(sum(sum(tort)))/(num*num*(num-1))
    im = np.swapaxes(im, 0, axis)
    return tort, tortavg



r"""

Visualization
#############

**Create Basic Views**

This module contains functions for quickly visualizing 3D images in 2D
views.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

   visualization.sem
   visualization.show_planes
   visualization.xray
   visualization.show_3D
   visualization.imshow
   visualization.bar
   visualization.show_mesh
   visualization.set_mpl_style

"""

__all__ = [
    "sem",
    "show_planes",
    "xray",
    "show_3D",
    "imshow",
    "bar",
    "show_mesh",
    "set_mpl_style"]

from ._views import sem, show_planes, xray, show_3D
from ._plots import imshow, bar, show_mesh
from ._funcs import set_mpl_style, satn_to_movie



import numpy as np
from loguru import logger


def imagej_wrapper(im, plugin_name, path):
    r"""
    Apply ImageJ filters on 3D images.

    Parameters
    ----------
    im : ndarray
        The 3D image of the porous material.

    plugin_name : str
        Name of the applied ImageJ plugin which could be one of these options:
            - Distance Transform Watershed 3D
            - Dilation 3D
            - Erosion 3D
            - Opening 3D
            - Closing 3D
            - Median 3D
            - Gaussian Blur 3D
            - Minimum 3D
            - Maximum 3D
            - Mean 3D

    path: str
        path to the Fiji application in the local directory

    Returns
    -------
    ndarray
        Outputs a ndarray after applying the desired filter on the image.

    """
    try:
        import imagej
        from scyjava import jimport
    except ModuleNotFoundError:
        msg = ("The pyimagej python bindings must be installed using conda"
               " install -c conda-forge paraview, however this may require"
               " using a virtualenv since conflicts with other packages are"
               " common. This is why it is not explicitly included as a"
               " dependency in porespy.")
        logger.critical(msg)

    i = False
    ij = imagej.init(path, headless=False)
    img = 255 * np.array(im.astype("uint8"))
    WindowManager = jimport("ij.WindowManager")
    ij.ui().show("Image", ij.py.to_java(img))
    plugin = "Duplicate..."
    args = {"duplicate range": (1, im.shape[0])}
    ij.py.run_plugin(plugin, args)

    if "watershed" in plugin_name.lower():
        ij.py.run_macro("""run("Distance Transform Watershed 3D","options=True");""")
        ij.py.run_macro("""run("3-3-2 RGB");""")
    elif "erosion" in plugin_name.lower():
        plugin = "Morphological Filters (3D)"
        args = {"operation": "Erosion"}
        ij.py.run_plugin(plugin, args)
    elif "dilation" in plugin_name.lower():
        plugin = "Morphological Filters (3D)"
        args = {"operation": "Dilation"}
        ij.py.run_plugin(plugin, args)
    elif "opening" in plugin_name.lower():
        plugin = "Morphological Filters (3D)"
        args = {"operation": "Opening"}
        ij.py.run_plugin(plugin, args)
    elif "closing" in plugin_name.lower():
        plugin = "Morphological Filters (3D)"
        args = {"operation": "Closing"}
        ij.py.run_plugin(plugin, args)
    elif "median" in plugin_name.lower():
        plugin = "Median 3D..."
        args = {"options": "True"}
        ij.py.run_plugin(plugin, args)
        i = True
    elif "mean" in plugin_name.lower():
        plugin = "Mean 3D..."
        args = {"options": "True"}
        ij.py.run_plugin(plugin, args)
        i = True
    elif "gaussian" in plugin_name.lower():
        plugin = "Gaussian Blur 3D..."
        args = {"options": "True"}
        ij.py.run_plugin(plugin, args)
        i = True
    elif "maximum" in plugin_name.lower():
        plugin = "Maximum 3D..."
        args = {"options": "True"}
        ij.py.run_plugin(plugin, args)
        i = True
    elif "minimum" in plugin_name.lower():
        plugin = "Minimum 3D..."
        args = {"options": "True"}
        ij.py.run_plugin(plugin, args)
        i = True
    results = np.array(ij.py.from_java(WindowManager.getCurrentImage()))
    if i:
        results = np.moveaxis(results, -1, 0)
    ij.getContext().dispose()
    WindowManager.closeAllWindows()
    return results


def imagej_plugin(im, path, plugin_name, args=None):
    r"""
    Apply ImageJ filters on 3D images.

    In This function the plugin_name should have a same format as the
    plugin_name in the ImageJ. For example, to apply a Gaussian blur on a 3D
    image, the plugin_name should be 'Gaussian Blur 3D...'

    Parameters
    ----------
    im : ndarray
        The 3D image of the porous material

    path: str
        Path to the Fiji application in the local directory

    plugin_name : str
        Name of the applied ImageJ plugin

    args : dict
         A dictionary that containes the required arguments of the
         applied plugin. For example, it could be {'options': 'True'}

    Returns
    -------
    ndarray
        Outputs a ndarray after applying the desired filter on the image.

    """
    try:
        import imagej
        from scyjava import jimport
    except ModuleNotFoundError:
        msg = ("The pyimagej python bindings must be installed using conda"
               " install -c conda-forge paraview, however this may require"
               " using a virtualenv since conflicts with other packages are"
               " common. This is why it is not explicitly included as a"
               " dependency in porespy.")
        logger.critical(msg)
    ij = imagej.init(path, headless=False)
    img = 255 * np.array(im.astype("uint8"))
    WindowManager = jimport('ij.WindowManager')
    ij.ui().show('Image', ij.py.to_java(img))
    plugin = 'Duplicate...'
    arg = {'duplicate range': (1, im.shape[0])}
    ij.py.run_plugin(plugin, arg)
    ij.py.run_plugin(plugin_name, args)
    results = np.array(ij.py.from_java(WindowManager.getCurrentImage()))
    ij.getContext().dispose()
    WindowManager.closeAllWindows()
    return (results)



import numpy as np
import openpnm as op
import scipy.ndimage as spim
from skimage.segmentation import find_boundaries
from skimage.morphology import ball, cube, disk, square
from porespy.tools import make_contiguous
from porespy.tools import overlay
from porespy.tools import insert_cylinder
from porespy.generators import borders
from porespy import settings
from porespy.tools import get_tqdm
from loguru import logger
tqdm = get_tqdm()


def map_to_regions(regions, values):
    r"""
    Maps pore values from a network onto the image from which it was extracted

    Parameters
    ----------
    regions : ndarray
        An image of the pore space partitioned into regions and labeled
    values : array_like
        An array containing the numerical values to insert into each region.
        The value at location *n* will be inserted into the image where
        ``regions`` is *n+1*.  This mismatch is caused by the fact that 0's
        in the ``regions`` image is assumed to be the backgroung phase, while
        pore index 0 is valid.

    Notes
    -----
    This function assumes that the array of pore values are indexed starting
    at location 0, while in the region image 0's indicate background phase and
    the region indexing starts at 1.  That is, region 1 corresponds to pore 0.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/networks/howtos/map_to_regions.html>`_
    to view online example.

    """
    values = np.array(values).flatten()
    if np.size(values) != regions.max():
        raise Exception('Number of values does not match number of regions')
    im = np.zeros_like(regions)
    im = values[regions-1]
    im = im*(regions > 0)
    return im


def add_boundary_regions(regions, pad_width=3):
    r"""
    Add boundary regions on specified faces of an image

    Parameters
    ----------
    regions : ndarray
        An image containing labelled regions, such as a watershed segmentation
    pad_width : array_like
        Number of layers to add to the beginning and end of each axis. This
        argument is handled similarly to the ``pad_width`` in the ``np.pad``
        function. A scalar adds the same amount to the beginning and end of
        each axis. [A, B] adds A to the beginning of each axis and B to the
        ends.  [[A, B], ..., [C, D]] adds A to the beginning and B to the
        end of the first axis, and so on. The default is to add 3 voxels on
        both ends of each axis.  One exception is is [A, B, C] which A to
        the beginning and end of the first axis, and so on. This extra option
        is useful for putting 0 on some axes (i.e. [3, 0, 0]).

    Returns
    -------
    padded_regions : ndarray
        An image with new regions padded on each side of the specified
        width.

    """
    # Parse user specified padding
    faces = np.array(pad_width)
    if faces.size == 1:
        faces = np.array([[faces, faces]]*regions.ndim)
    elif faces.size == regions.ndim:
        faces = np.vstack([faces]*2)
    else:
        pass
    t = faces.max()
    mx = regions.max()
    # Put a border around each region so padded regions are isolated
    bd = find_boundaries(regions, connectivity=regions.ndim, mode='inner')
    # Pad by t in all directions, this will be trimmed down later
    face_regions = np.pad(regions*(~bd), pad_width=t, mode='edge')
    # Set corners to 0 so regions don't connect across faces
    edges = borders(shape=face_regions.shape, mode='edges', thickness=t)
    face_regions[edges] = 0
    # Extract a mask of just the faces
    mask = borders(shape=face_regions.shape, mode='faces', thickness=t)
    # Relabel regions on faces
    new_regions = spim.label(face_regions*mask)[0] + mx*(face_regions > 0)
    new_regions[~mask] = regions.flatten()
    # Trim image down to user specified size
    s = tuple([slice(t-ax[0], -(t-ax[1]) or None) for ax in faces])
    new_regions = new_regions[s]
    new_regions = make_contiguous(new_regions)
    return new_regions


def _generate_voxel_image(network, pore_shape, throat_shape, max_dim=200):
    r"""
    Generates a 3d numpy array from an OpenPNM network

    Parameters
    ----------
    network : OpenPNM GenericNetwork
        Network from which voxel image is to be generated
    pore_shape : str
        Shape of pores in the network, valid choices are "sphere", "cube"
    throat_shape : str
        Shape of throats in the network, valid choices are "cylinder", "cuboid"
    max_dim : int
        Number of voxels in the largest dimension of the network

    Returns
    -------
    im : ndarray
        Voxelated image corresponding to the given pore network model

    Notes
    -----
    (1) The generated voxel image is labeled with 0s, 1s and 2s signifying
    solid phase, pores, and throats respectively.

    """
    xyz = network["pore.coords"]
    cn = network["throat.conns"]

    # Distance bounding box from the network by a fixed amount
    delta = network["pore.diameter"].mean() / 2
    if isinstance(network, op.network.Cubic):
        delta = network._spacing.mean() / 2

    # Shift everything to avoid out-of-bounds
    extra_clearance = int(max_dim * 0.05)

    # Transform points to satisfy origin at (0, 0, 0)
    xyz0 = xyz.min(axis=0) - delta
    xyz += -xyz0
    res = (xyz.ptp(axis=0).max() + 2 * delta) / max_dim
    shape = np.rint((xyz.max(axis=0) + delta) / res).astype(int) + 2 * extra_clearance

    # Transforming from real coords to matrix coords
    xyz = np.rint(xyz / res).astype(int) + extra_clearance
    pore_radi = np.rint(network["pore.diameter"] * 0.5 / res).astype(int)
    throat_radi = np.rint(network["throat.diameter"] * 0.5 / res).astype(int)

    im_pores = np.zeros(shape, dtype=np.uint8)
    im_throats = np.zeros_like(im_pores)

    if pore_shape == "cube":
        pore_elem = cube
        rp = pore_radi * 2 + 1  # +1 since num_voxel must be odd
        rp_max = int(2 * round(delta / res)) + 1
    if pore_shape == "sphere":
        pore_elem = ball
        rp = pore_radi
        rp_max = int(round(delta / res))
    if throat_shape == "cuboid":
        raise Exception("Not yet implemented, try 'cylinder'.")

    # Generating voxels for pores
    for i, pore in enumerate(tqdm(network.Ps, **settings.tqdm)):
        elem = pore_elem(rp[i])
        try:
            im_pores = overlay(im1=im_pores, im2=elem, c=xyz[i])
        except ValueError:
            elem = pore_elem(rp_max)
            im_pores = overlay(im1=im_pores, im2=elem, c=xyz[i])
    # Get rid of pore overlaps
    im_pores[im_pores > 0] = 1

    # Generating voxels for throats
    for i, throat in enumerate(tqdm(network.Ts, **settings.tqdm)):
        try:
            im_throats = insert_cylinder(
                im_throats, r=throat_radi[i], xyz0=xyz[cn[i, 0]], xyz1=xyz[cn[i, 1]])
        except ValueError:
            im_throats = insert_cylinder(
                im_throats, r=rp_max, xyz0=xyz[cn[i, 0]], xyz1=xyz[cn[i, 1]])
    # Get rid of throat overlaps
    im_throats[im_throats > 0] = 1

    # Subtract pore-throat overlap from throats
    im_throats = (im_throats.astype(bool) * ~im_pores.astype(bool)).astype(np.uint8)
    im = im_pores * 1 + im_throats * 2

    return im[extra_clearance:-extra_clearance,
              extra_clearance:-extra_clearance,
              extra_clearance:-extra_clearance]


def generate_voxel_image(network, pore_shape="sphere", throat_shape="cylinder",
                         max_dim=None, rtol=0.1):
    r"""
    Generate a voxel image from an OpenPNM network object

    Parameters
    ----------
    network : OpenPNM GenericNetwork
        Network from which voxel image is to be generated
    pore_shape : str
        Shape of pores in the network, valid choices are "sphere", "cube"
    throat_shape : str
        Shape of throats in the network, valid choices are "cylinder", "cuboid"
    max_dim : int
        Number of voxels in the largest dimension of the network
    rtol : float
        Stopping criteria for finding the smallest voxel image such that
        further increasing the number of voxels in each dimension by 25% would
        improve the predicted porosity of the image by less that ``rtol``

    Returns
    -------
    im : ndarray
        Voxelated image corresponding to the given pore network model

    Notes
    -----
    (1) The generated voxelated image is labeled with 0s, 1s and 2s signifying
    solid phase, pores, and throats respectively.

    (2) If max_dim is not provided, the method calculates it such that the
    further increasing it doesn't change porosity by much.

    """
    logger.trace("Generating voxel image from pore network")
    # If max_dim is provided, generate voxel image using max_dim
    if max_dim is not None:
        return _generate_voxel_image(
            network, pore_shape, throat_shape, max_dim=max_dim)
    max_dim = 200
    # If max_dim is not provided, find best max_dim that predicts porosity
    err = 100
    eps_old = 200
    while err > rtol:
        logger.debug(f"Maximum dimension: {max_dim} voxels")
        im = _generate_voxel_image(
            network, pore_shape, throat_shape, max_dim=max_dim)
        eps = im.astype(bool).sum() / np.prod(im.shape)
        err = abs(1 - eps / eps_old)
        eps_old = eps
        max_dim = int(max_dim * 1.25)
    logger.debug(f"Converged at max_dim = {max_dim} voxels")
    return im


def label_phases(
        network,
        alias={1: 'void', 2: 'solid'}):
    r"""
    Create pore and throat labels based on 'pore.phase' values

    Parameters
    ----------
    network : dict
        The network stored as a dictionary as returned from the
        ``regions_to_network`` function
    alias : dict
        A mapping between integer values in 'pore.phase' and string labels.
        The default is ``{1: 'void', 2: 'solid'}`` which will result in the
        labels ``'pore.void'`` and ``'pore.solid'``, as well as
        ``'throat.solid_void'``, ``'throat.solid_solid'``, and
        ``'throat.void_void'``.  The reverse labels are also added for
        convenience like ``throat.void_solid``.

    Returns
    -------
    network : dict
        The same ``network`` as passed in but with new boolean arrays added
        for the phase labels.

    """
    conns = network['throat.conns']
    for i in alias.keys():
        pore_i_hits = network['pore.phase'] == i
        network['pore.' + alias[i]] = pore_i_hits
        for j in alias.keys():
            pore_j_hits = network['pore.phase'] == j
            throat_hits = pore_i_hits[conns[:, 0]] * pore_j_hits[conns[:, 1]]
            throat_hits += pore_i_hits[conns[:, 1]] * pore_j_hits[conns[:, 0]]
            if np.any(throat_hits):
                name = 'throat.' + '_'.join([alias[i], alias[j]])
                if name not in network.keys():
                    network[name] = np.zeros_like(conns[:, 0], dtype=bool)
                network[name] += throat_hits
    return network


def label_boundaries(
        network,
        labels=[['left', 'right'], ['front', 'back'], ['top', 'bottom']],
        tol=1e-9):
    r"""
    Create boundary pore labels based on proximity to axis extrema

    Parameters
    ----------
    network : dict
        The network stored as a dictionary as returned from the
        ``regions_to_network`` function
    labels : list of lists
        A 3-element list, with each element containing a pair of strings
        indicating the label to apply to the beginning and end of each axis.
        The default is ``[['left', 'right'], ['front', 'back'],
        ['top', 'bottom']]`` which will apply the label ``'left'`` to all
        pores with the minimum x-coordinate, and ``'right'`` to the pores
        with the maximum x-coordinate, and so on.

    Returns
    -------
    network : dict
        The same ``network`` as passed in but with new boolean arrays added
        for the boundary labels.

    """
    crds = network['pore.coords']
    extents = [[crds[:, i].min(), crds[:, i].max()]
               for i in range(len(crds[0, :]))]
    network['pore.boundary'] = np.zeros_like(crds[:, 0], dtype=bool)
    for i, axis in enumerate(labels):
        for j, face in enumerate(axis):
            if face:
                hits = crds[:, i] == extents[i][j]
                network['pore.boundary'] += hits
                network['pore.' + labels[i][j]] = hits
    return network



r"""

Generators
##########

**Generate Artificial Images**

This module contains a variety of functions for generating artificial
images of porous materials, generally for testing, validation,
debugging, and illustration purposes.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    generators.blobs
    generators.bundle_of_tubes
    generators.cylindrical_plug
    generators.cylinders
    generators.fractal_noise
    generators.insert_shape
    generators.lattice_spheres
    generators.line_segment
    generators.overlapping_spheres
    generators.polydisperse_spheres
    generators.pseudo_electrostatic_packing
    generators.pseudo_gravity_packing
    generators.RSA
    generators.random_cantor_dust
    generators.sierpinski_foam
    generators.voronoi_edges

"""

from ._imgen import blobs
from ._imgen import bundle_of_tubes
from ._imgen import cylinders
from ._imgen import insert_shape
from ._imgen import lattice_spheres
from ._imgen import line_segment
from ._imgen import overlapping_spheres
from ._imgen import polydisperse_spheres
from ._imgen import RSA
from ._imgen import voronoi_edges
from ._pseudo_packings import pseudo_gravity_packing
from ._pseudo_packings import pseudo_electrostatic_packing
from ._cylinder import cylindrical_plug
from ._noise import fractal_noise
from ._borders import *
from ._fractals import random_cantor_dust
from ._fractals import sierpinski_foam
